#!/usr/bin/env python
#
# chisel.py
# Uses spherical needlets to estimate a selection function across the sky.
#
# Copyright (C) 2021  Douglas Boubert & Andrew Everall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import numpy as np
import healpy as hp
import tqdm
import h5py
import copy
from .base import Base

from PythonModels.wavelet_magnitude_colour_position import wavelet_magnitude_colour_position_sparse, wavelet_x_sparse, wavelet_b_sparse

class Chisel(Base):

    basis_keyword = 'wavelet'

    def _process_basis_options(self, needlet = 'littlewoodpaley', j=[0], B = 2.0, p = 1.0, wavelet_tol = 1e-10):

        if type(j) in [list,tuple,np.ndarray]:
            self.j = sorted([int(_j) for _j in j])
        else:
            self.j = [_j for _j in range(-1,j+1)]
        self.needlet, self.B, self.p, self.wavelet_tol = needlet, B, p, wavelet_tol

        assert self.B > 1.0
        assert self.wavelet_tol >= 0.0
        assert self.needlet in ['littlewoodpaley','chisquare']
        self.S = sum([self.order_to_npix(_j) if _j >= 0 else 1 for _j in self.j])
        self.nside = hp.npix2nside(self.P)

        self.spherical_basis_file = f"{self.basis_keyword}_{self.needlet}_nside{self.nside}_B{self.B}_"+ (f"p{self.p}_" if self.needlet == 'chisquare' else '') + f"tol{self.wavelet_tol}_j[{','.join([str(_i) for _i in self.j])}].h5"

        if self.needlet == 'chisquare':
            from .utils import chisquare
            self.weighting = chisquare(self.j, p = self.p, B = self.B)
        else:
            from .utils import littlewoodpaley
            self.weighting = littlewoodpaley(B = self.B)

    def _process_sigma_basis_specific(self,sigma):
        assert len(sigma) == 2
        power_spectrum = lambda l: np.exp(sigma[0]) * np.power(1.0+l,sigma[1])

        _sigma = np.zeros(self.S)
        running_index = 0
        for j in self.j:

            if j == -1:
                _sigma[running_index] = np.exp(sigma[0])
                running_index += 1
                continue

            npix_needle = self.order_to_npix(j)

            start = self.weighting.start(j)
            end = self.weighting.end(j)
            modes = np.arange(start, end + 1, dtype = 'float')
            _lambda = 4*np.pi/npix_needle # 1/np.sum(self.weighting.window_function(modes,j)* (2.0*modes+1.0)/(4*np.pi))**2 #1/npix_needle
            window = _lambda * self.weighting.window_function(modes,j)**2 * (2.0*modes+1.0)/(4*np.pi) * power_spectrum(modes)

            _sigma[running_index:running_index+npix_needle] = np.sqrt(window.sum())
            running_index += npix_needle

        # Renormalise (Marinucci 2008)
        #l = np.arange(1,self.weighting.lmax)
        #print(np.sum(power_spectrum((2*l+1)/4*np.pi * power_spectrum(l))) / np.sum(_sigma**2))
        #print(np.sum(power_spectrum((2*l+1)/4*np.pi * power_spectrum(l))) / np.sum(_sigma[1:]**2))
        #_sigma[0] = 6.9

        return _sigma

    def _generate_spherical_basis(self,gsb_file, coords=None):

        # Import dependencies
        from numba import njit
        from math import sin, cos
        import sys

        nside = self.nside
        B = self.B
        needle_sparse_tol = self.wavelet_tol

        # Function to compute needlet across sky
        @njit
        def pixel_space (Y, cos_gamma, window, start, end, legendre):
            '''Return the value of a needlet at gamma radians from the needlet centre.'''

            legendre[0] = 1.0
            legendre[1] = cos_gamma
            for cur_l in range(2, end + 1):
                legendre[cur_l] = ((cos_gamma * (2 * cur_l - 1) * legendre[cur_l - 1] - (cur_l - 1) * legendre[cur_l - 2])) / cur_l

            Y[:] = np.dot(window,legendre[start:end+1])

        # Compute locations of pixels
        if coords is None:
            npix = self.nside_to_npix(nside)
            colat, lon = np.array(hp.pix2ang(nside=nside,ipix=np.arange(npix),lonlat=False))
        else:
            colat, lon = coords
            npix = len(colat)
        cos_colat, sin_colat = np.cos(colat), np.sin(colat)
        cos_lon, sin_lon = np.cos(lon), np.sin(lon)

        # Initialise variables
        running_index = 0
        needlet_w, needlet_v, needlet_u, needlet_un, needlet_j = [], [], [], [], []
        Y = np.zeros(npix)
        legendre = np.zeros((1+self.weighting.end(max(self.j)),npix))

        for ineedlet, j in enumerate(self.j):

            print(f'Working on order {j}.')

            if j == -1:
                needlet_w.append(np.ones(npix))
                needlet_v.append(np.arange(npix))
                needlet_u.append(0)
                needlet_un.append(np.zeros(npix, dtype=np.uint64))
                needlet_j.append(np.zeros(1))
                running_index += npix
                continue

            nside_needle = self.order_to_nside(j)
            npix_needle = self.nside_to_npix(nside_needle)

            start = self.weighting.start(j)
            end = self.weighting.end(j)
            modes = np.arange(start, end + 1, dtype = 'float')
            _lambda = 4*np.pi/npix_needle # 1/np.sum(self.weighting.window_function(modes,j)* (2.0*modes+1.0)/(4*np.pi))**2 # 1/npix_needle
            window = np.sqrt(_lambda) * self.weighting.window_function(modes,j) * (2.0*modes+1.0)/(4.0*np.pi)

            for ipix_needle in tqdm.tqdm(range(npix_needle),file=sys.stdout):

                colat_needle, lon_needle = hp.pix2ang(nside=nside_needle,ipix=ipix_needle,lonlat=False)

                cos_gamma = cos(colat_needle) * cos_colat + sin(colat_needle) * sin_colat * (cos(lon_needle) * cos_lon + sin(lon_needle) * sin_lon)

                pixel_space(Y, cos_gamma = cos_gamma, window = window, start = start, end = end, legendre = legendre)

                _significant = np.where(np.abs(Y) > Y.max()*needle_sparse_tol)[0]
                needlet_w.append(Y[_significant])
                needlet_v.append(_significant)
                needlet_u.append(running_index)
                needlet_un.append(np.zeros(_significant.size, dtype=np.uint64) +j+ipix_needle )
                needlet_j.append(j*np.ones(self.order_to_npix(j)))
                running_index += _significant.size

        # Add the ending index to u
        needlet_u.append(running_index)

        # Concatenate the lists
        needlet_w = np.concatenate(needlet_w)
        needlet_v = np.concatenate(needlet_v)
        needlet_u = np.array(needlet_u)

        # Flip them round
        from scipy import sparse
        Y = sparse.csr_matrix((needlet_w,needlet_v,needlet_u)).transpose().tocsr()
        wavelet_w, wavelet_v, wavelet_u = Y.data, Y.indices, Y.indptr
        wavelet_j = np.concatenate(needlet_j).astype(int)
        wavelet_n = wavelet_w.size

        print('Expanding u')
        @njit
        def expand_u(wavelet_u, wavelet_U):
            size = wavelet_u.size-1
            for iS in range(size):
                wavelet_U[wavelet_u[iS]:wavelet_u[iS+1]] = iS
        wavelet_U = np.zeros(wavelet_v.size, dtype=np.uint64)
        expand_u(wavelet_u, wavelet_U)

        if coords is None:
            # Save file
            save_kwargs = {'compression':"lzf", 'chunks':True, 'fletcher32':False, 'shuffle':True}
            with h5py.File(gsb_file, 'w') as f:
                f.create_dataset('wavelet_w', data = wavelet_w, dtype = np.float64, **save_kwargs)
                f.create_dataset('wavelet_v', data = wavelet_v, dtype = np.uint64, scaleoffset=0, **save_kwargs)
                f.create_dataset('wavelet_u', data = wavelet_u, dtype = np.uint64, scaleoffset=0, **save_kwargs)
                f.create_dataset('wavelet_U', data = wavelet_U, dtype = np.uint64, scaleoffset=0, **save_kwargs)
                f.create_dataset('wavelet_n', data = wavelet_n)
                f.create_dataset('modes', data = wavelet_j, dtype = np.uint64, scaleoffset=0, **save_kwargs)
        else:
            return Y

    def _yield_stan_model(self):
        if self.sparse:
            cholesky_parameters = '''
            int cholesky_n_m;                     // sparse cholesky in magnitude - number of nonzero elements
            row_vector[cholesky_n_m] cholesky_w_m;// sparse cholesky in magnitude - nonzero elements
            int cholesky_v_m[cholesky_n_m];       // sparse cholesky in magnitude - columns of nonzero elements
            int cholesky_u_m[M+1];                // sparse cholesky in magnitude - where in w each row starts
            int cholesky_n_c;                     // sparse cholesky in colour - number of nonzero elements
            vector[cholesky_n_c] cholesky_w_c;    // sparse cholesky in colour - nonzero elements
            int cholesky_v_c[cholesky_n_c];       // sparse cholesky in colour - columns of nonzero elements
            int cholesky_u_c[C+1];                // sparse cholesky in colour - where in w each row starts
            '''
            cholesky_loop = '''
            b[s] = sigma[s] * (cholesky_w_m[cholesky_u_m[m]:cholesky_u_m[m+1]-1] * z[s,cholesky_v_m[cholesky_u_m[m]:cholesky_u_m[m+1]-1], cholesky_v_c[cholesky_u_c[c]:cholesky_u_c[c+1]-1]] * cholesky_w_c[cholesky_u_c[c]:cholesky_u_c[c+1]-1]);
            '''
        else:
            cholesky_parameters = '''
            row_vector[M_subspace] cholesky_m[M]; // Cholesky factor in magnitude space
            vector[C_subspace] cholesky_c[C];     // Cholesky factor in colour space
            '''
            cholesky_loop = '''
            b[s] = sigma[s] * cholesky_m[m] * z[s] * cholesky_c[c];
            '''

        stan_model = f'''
        data {{
            int<lower=0> P;                       // number of pixels
            int<lower=0> M;                       // number of bins in magnitude space
            int<lower=0> M_subspace;              // number of inducing points in magnitude space
            int<lower=0> C;                       // number of bins in colour space
            int<lower=0> C_subspace;              // number of inducing points in colour space
            int<lower=0> S;                       // number of wavelets
            int wavelet_n;                        // sparse wavelets - number of nonzero elements
            vector[wavelet_n] wavelet_w;          // sparse wavelets - nonzero elements
            int wavelet_v[wavelet_n];             // sparse wavelets - columns of nonzero elements
            int wavelet_u[P+1];                   // sparse wavelets - where in w each row starts
            real mu;                              // mean across sky
            vector[S] sigma;                      // sigma of each wavelet
            int k[M,C,P];                         // number of heads
            int n[M,C,P];                         // number of flips
            {cholesky_parameters}
        }}
        parameters {{
            matrix[M_subspace,C_subspace] z[S];
        }}
        transformed parameters {{

            vector[P] x[M,C]; // Probability in logit-space

            // Loop over magnitude and colour
            for (m in 1:M){{
                for (c in 1:C){{

                    // Local variable
                    vector[S] b;

                    // Compute b
                    for (s in 1:S){{
                        {cholesky_loop}
                    }}

                    // Compute x
                    x[m,c] = mu + csr_matrix_times_vector(P, S, wavelet_w, wavelet_v, wavelet_u, b);

                }}
            }}

        }}
        model {{

            // Prior
            for (s in 1:S){{
                to_vector(z[s]) ~ std_normal();
            }}

            // Likelihood
            for (m in 1:M){{
                for (c in 1:C){{
                    k[m,c] ~ binomial_logit(n[m,c], x[m,c]);
                }}
            }}

        }}
        '''

        return stan_model

    def _cholesky_args(self, sparse=False):

        if sparse:
            cholesky_u_m = np.zeros(len(self.model_input['cholesky_v_m']), dtype=int)
            for iS, iY in enumerate(self.model_input['cholesky_u_m'][1:]-1):
                cholesky_u_m[iY:] += 1
            cholesky_u_c = np.zeros(len(self.model_input['cholesky_v_c']), dtype=int)
            for iS, iY in enumerate(self.model_input['cholesky_u_c'][1:]-1):
                cholesky_u_c[iY:] += 1

            cholesky_args = [cholesky_u_m,
                             self.model_input['cholesky_v_m']-1,
                             self.model_input['cholesky_w_m'],
                             cholesky_u_c,
                             self.model_input['cholesky_v_c']-1,
                             self.model_input['cholesky_w_c']]

        elif not sparse:
            cholesky_args = [self.cholesky_m, self.cholesky_c]

        return cholesky_args

    def _construct_scipy_input(self, sparse=False):

        if not sparse: raise ValueError(f'Sparse={sparse}. Only sparse model implemented for Chisel with scipy.')

        cholesky_args = self._cholesky_args(sparse=sparse)

        lnL_grad = np.zeros((self.S, self.M, self.C))
        x = np.zeros((self.M, self.C))

        self.scipy_args = [np.moveaxis(self.k, -1,0).astype(np.int64).copy(),
                           np.moveaxis(self.n, -1,0).astype(np.int64).copy()] \
                          + [np.zeros(self.S)+copy.copy(self.model_input[arg]) for arg in ['mu', 'sigma']]\
                          + [self.model_input['wavelet_U'].copy()-1,
                             self.model_input['wavelet_v'].copy()-1,
                             self.model_input['wavelet_w'].copy()]\
                          + cholesky_args

        if sparse: self.likelihood_function = wavelet_magnitude_colour_position_sparse

    def _parallelize_scipy_model(self, nsets=1):

        P_set = np.zeros(nsets, dtype=int) + self.P//nsets
        P_set[:self.P - np.sum(P_set)] += 1
        print('P sets: ', P_set, self.P)

        wavelet_u = self.model_input['wavelet_U'].copy()-1
        wavelet_v = self.model_input['wavelet_v'].copy()-1
        wavelet_w = self.model_input['wavelet_w'].copy()

        self.scipy_args_set = []
        iP = 0
        for iset in range(nsets):

            lnL_grad = np.zeros((self.S, self.M, self.C))
            x = np.zeros((self.M, self.C))

            scipy_args_set  = [self.scipy_args[i][iP:iP+P_set[iset]].copy() for i in [0,1]] \
                              + [self.scipy_args[i] for i in [2,3]] \
                              + [self.scipy_args[i][int(self.model_input['wavelet_u'][iP]-1):int(self.model_input['wavelet_u'][iP+P_set[iset]]-1)].copy() \
                                                                        for i in [4,5,6]] \
                              + [self.scipy_args[i] for i in range(7,len(self.scipy_args))]
            scipy_args_set[4]-=iP

            self.scipy_args_set.append(scipy_args_set)
            iP += P_set[iset]

        self.P_set = P_set

    def _get_b(self, z):

        from numba.typed import List

        if not hasattr(self, 'scipy_args'): self._load_scipy_model(sparse=self.sparse)

        print('Setting cholesky arguments')
        cholesky_m = List(zip(*self.scipy_args[7:10]))
        cholesky_c = List(zip(*self.scipy_args[10:13]))

        b = np.zeros((self.S, self.M, self.C))
        b = wavelet_b_sparse(z.reshape((self.S, self.M_subspace, self.C_subspace)), self.M, self.C, self.S, *self.scipy_args[2:4], cholesky_m, cholesky_c, b)

        return b

    def _get_x(self, z):

        from numba.typed import List

        if not hasattr(self, 'scipy_args'): self._load_scipy_model(sparse=self.sparse)

        print('Setting cholesky arguments')
        wavelet = List(zip(*self.scipy_args[4:7]))
        cholesky_m = List(zip(*self.scipy_args[7:10]))
        cholesky_c = List(zip(*self.scipy_args[10:13]))

        x = np.zeros((self.P, self.M, self.C))
        x = wavelet_x_sparse(z.reshape((self.S, self.M_subspace, self.C_subspace)), self.M, self.C, *self.scipy_args[2:4], wavelet, cholesky_m, cholesky_c, x)

        return x



#!/usr/bin/env python
#
# kernel.py
# Covariance kernels for the magnitude and colour dimensions.
#
# Copyright (C) 2021  Douglas Boubert & Andrew Everall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import numpy as np
import copy

class kernelbase:

    def __init__(self, variance = 1.0, pivot = False, pivot_tol = 0.0, eps_tol = 1e-15, **kwargs):
        self.variance = variance
        self.pivot = pivot
        self.pivot_tol = pivot_tol
        self.eps_tol = eps_tol
        self._process_kernel_arguments(**kwargs)

    def distance(self, x1, x2):
        return np.abs(x1[:,np.newaxis]-x2[np.newaxis,:])

    def covariance(self, x1, x2):
        pass

    def cholesky(self, x):

        _covariance = self.covariance(x, x)
        _n = x.shape[0]

        if self.pivot:
            _cholesky = self._pivoted_cholesky(_covariance, M=_n, err_tol=self.pivot_tol)
        else:
            _cholesky = np.linalg.cholesky(_covariance+self.eps_tol*np.diag(np.ones(_n)))

        _n_subspace = _cholesky.shape[1]

        return _n_subspace, _cholesky

    def __add__(self, other):

        if isinstance(other,(float,int)):
            _new_kernel = AdditiveKernel(kernel1=self,kernel2=Flat(variance=other))
        elif isinstance(other,kernelbase):
            _new_kernel = AdditiveKernel(kernel1=self,kernel2=other)
        return _new_kernel

    __radd__ = __add__

    def __mul__(self, other):

        if isinstance(other,(float,int)):
            _new_kernel = copy.deepcopy(self)
            _new_kernel.variance = self.variance*other
        elif isinstance(other,kernelbase):
            _new_kernel = MultiplicativeKernel(kernel1=self,kernel2=other)
        return _new_kernel

    __rmul__ = __mul__


    def _process_kernel_arguments(self):
        pass

    def _pivoted_cholesky(self, A, M, err_tol = 1e-6):
        """
        https://dl.acm.org/doi/10.1016/j.apnum.2011.10.001 implemented by https://github.com/NathanWycoff/PivotedCholesky
        A simple python function which computes the Pivoted Cholesky decomposition/approximation of positive semi-definite operator. Only diagonal elements and select rows of that operator's matrix represenation are required.
        get_diag - A function which takes no arguments and returns the diagonal of the matrix when called.
        get_row - A function which takes 1 integer argument and returns the desired row (zero indexed).
        M - The maximum rank of the approximate decomposition; an integer.
        err_tol - The maximum error tolerance, that is difference between the approximate decomposition and true matrix, allowed. Note that this is in the Trace norm, not the spectral or frobenius norm.
        Returns: R, an upper triangular matrix of column dimension equal to the target matrix. It's row dimension will be at most M, but may be less if the termination condition was acceptably low error rather than max iters reached.
        """

        get_diag = lambda: np.diag(A).copy()
        get_row = lambda i: A[i,:]

        d = np.copy(get_diag())
        N = len(d)

        pi = list(range(N))

        R = np.zeros([M,N])

        err = np.sum(np.abs(d))

        m = 0
        while (m < M) and (err > err_tol):

            i = m + np.argmax([d[pi[j]] for j in range(m,N)])

            tmp = pi[m]
            pi[m] = pi[i]
            pi[i] = tmp

            R[m,pi[m]] = np.sqrt(d[pi[m]])
            Apim = get_row(pi[m])
            for i in range(m+1, N):
                if m > 0:
                    ip = np.inner(R[:m,pi[m]], R[:m,pi[i]])
                else:
                    ip = 0
                R[m,pi[i]] = (Apim[pi[i]] - ip) / R[m,pi[m]]
                d[pi[i]] -= pow(R[m,pi[i]],2)

            err = np.sum([d[pi[i]] for i in range(m+1,N)])
            m += 1

        R = R[:m,:]

        return(R.T)

class AdditiveKernel(kernelbase):

    def covariance(self, x1, x2):
        return self.variance*(self.kernel1.covariance(x1, x2)+self.kernel2.covariance(x1, x2))

    def _process_kernel_arguments(self, kernel1, kernel2):
        self.kernel1 = copy.deepcopy(kernel1)
        self.kernel2 = copy.deepcopy(kernel2)
        self.pivot = kernel1.pivot or kernel2.pivot
        self.pivot_tol = max(kernel1.pivot_tol,kernel2.pivot_tol)

class MultiplicativeKernel(kernelbase):

    def covariance(self, x1, x2):
        return self.variance*self.kernel1.covariance(x1, x2)*self.kernel2.covariance(x1, x2)

    def _process_kernel_arguments(self, kernel1, kernel2):
        self.kernel1 = copy.deepcopy(kernel1)
        self.kernel2 = copy.deepcopy(kernel2)
        self.pivot = kernel1.pivot or kernel2.pivot
        self.pivot_tol = max(kernel1.pivot_tol,kernel2.pivot_tol)

class SquaredExponential(kernelbase):

    def covariance(self, x1, x2):
        distance = self.distance(x1, x2)
        return self.variance*np.exp(-0.5*np.square(distance/self.lengthscale))

    def _process_kernel_arguments(self, lengthscale = 1.0):
        self.lengthscale = lengthscale

class Matern12(kernelbase):

    def covariance(self, x1, x2):
        distance = self.distance(x1, x2)
        return self.variance*np.exp(-distance/self.lengthscale)

    def _process_kernel_arguments(self, lengthscale = 1.0):
        self.lengthscale = lengthscale

class Matern32(kernelbase):

    def covariance(self, x1, x2):
        distance = self.distance(x1, x2)
        u = np.sqrt(3.0)*distance/self.lengthscale
        return self.variance*np.exp(-u)*(1.0+u)

    def _process_kernel_arguments(self, lengthscale = 1.0):
        self.lengthscale = lengthscale

class RationalQuadratic(kernelbase):

    def covariance(self, x1, x2):
        distance = self.distance(x1, x2)
        return self.variance*np.power(1.0+0.5*np.square(distance/self.lengthscale)/self.mixturescale,-self.mixturescale)

    def _process_kernel_arguments(self, lengthscale = 1.0, mixturescale = 1.0):
        self.lengthscale = lengthscale
        self.mixturescale = mixturescale

class Periodic(kernelbase):

    def covariance(self, x1, x2):
        distance = self.distance(x1, x2)
        return self.variance*np.exp(-2.0*np.square(np.sin(np.pi*distance/self.period)/self.lengthscale))

    def _process_kernel_arguments(self, lengthscale = 1.0, period = 1.0):
        self.lengthscale = lengthscale
        self.period = period

class Linear(kernelbase):

    def covariance(self, x1, x2):
        return self.variance*(x1[:,np.newaxis]-self.intercept)*(x2[np.newaxis,:]-self.intercept)

    def _process_kernel_arguments(self, intercept = 0.0):
        self.intercept = intercept

class WhiteNoise(kernelbase):

    def covariance(self, x1, x2):
        distance = self.distance(x1, x2)
        covariance = np.zeros_like(distance)
        covariance[np.isclose(distance,0.0)] = self.variance
        return covariance

class Flat(kernelbase):

    def covariance(self, x1, x2):
        return self.variance*np.ones((x1.size,x2.size))


#!/usr/bin/env python
#
# Models a selection function that is independent between positions.
#
# Copyright (C) 2021  Douglas Boubert & Andrew Everall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import numpy as np
import healpy as hp
import tqdm
import h5py
from .base import Base
import os

class Wrench(Base):

    basis_keyword = 'basic'

    def _process_basis_options(self):
        self.S = self.P
        self.spherical_basis_file = f'{self.basis_keyword}.h5'

    def _generate_spherical_basis(self,gsb_file):
        with h5py.File(gsb_file, 'w') as f:
            pass

    def _yield_stan_model(self):
        if self.sparse:
            cholesky_parameters = '''
            int cholesky_n_m;                     // sparse cholesky in magnitude - number of nonzero elements
            row_vector[cholesky_n_m] cholesky_w_m;// sparse cholesky in magnitude - nonzero elements
            int cholesky_v_m[cholesky_n_m];       // sparse cholesky in magnitude - columns of nonzero elements
            int cholesky_u_m[M+1];                // sparse cholesky in magnitude - where in w each row starts
            int cholesky_n_c;                     // sparse cholesky in colour - number of nonzero elements
            vector[cholesky_n_c] cholesky_w_c;    // sparse cholesky in colour - nonzero elements
            int cholesky_v_c[cholesky_n_c];       // sparse cholesky in colour - columns of nonzero elements
            int cholesky_u_c[C+1];                // sparse cholesky in colour - where in w each row starts
            '''
            cholesky_loop = '''
            x[m,c,s] = mu + sigma[s] * (cholesky_w_m[cholesky_u_m[m]:cholesky_u_m[m+1]-1] * z[s,cholesky_v_m[cholesky_u_m[m]:cholesky_u_m[m+1]-1], cholesky_v_c[cholesky_u_c[c]:cholesky_u_c[c+1]-1]] * cholesky_w_c[cholesky_u_c[c]:cholesky_u_c[c+1]-1]);
            '''
        else:
            cholesky_parameters = '''
            row_vector[M_subspace] cholesky_m[M]; // Cholesky factor in magnitude space
            vector[C_subspace] cholesky_c[C];     // Cholesky factor in colour space
            '''
            cholesky_loop = '''
            x[m,c,s] = mu + sigma[s] * cholesky_m[m] * z[s] * cholesky_c[c];
            '''

        stan_model = f'''
        data {{
            int<lower=0> P;                       // number of pixels
            int<lower=0> M;                       // number of bins in magnitude space
            int<lower=0> M_subspace;              // number of inducing points in magnitude space
            int<lower=0> C;                       // number of bins in colour space
            int<lower=0> C_subspace;              // number of inducing points in colour space
            int<lower=0> S;                       // number of modes
            real mu;                              // mean across sky
            vector[S] sigma;                      // sigma of each mode
            int k[M,C,P];                         // number of heads
            int n[M,C,P];                         // number of flips
            {cholesky_parameters}
        }}
        parameters {{
            matrix[M_subspace,C_subspace] z[S];
        }}
        transformed parameters {{

            vector[P] x[M,C]; // Probability in logit-space
                
            // Loop over magnitude and colour
            for (m in 1:M){{
                for (c in 1:C){{
                    // Compute x
                    for (s in 1:S){{
                        {cholesky_loop}
                    }}
                }}  
            }}

        }}
        model {{

            // Prior
            for (s in 1:S){{
                to_vector(z[s]) ~ std_normal();
            }}
            
            // Likelihood
            for (m in 1:M){{
                for (c in 1:C){{
                    k[m,c] ~ binomial_logit(n[m,c], x[m,c]);
                }}
            }}
            
        }}
        '''

        return stan_model


#!/usr/bin/env python
#
# base.py
# The base class of all tools in SelectionFunctionToolbox.
#
# Copyright (C) 2021  Douglas Boubert & Andrew Everall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import numpy as np
import healpy as hp
import tqdm
import h5py
import os
from .kernel import WhiteNoise
import copy, time

class Base:

    basis_keyword = 'base' # This must be changed in each file.

    def __init__(self, k, n, file_root, axes  = ['magnitude','colour','position'],
                        magnitude_bins = None, colour_bins = None, magnitude_kernel = None, colour_kernel = None,
                        sparse = False, sparse_tol = 1e-4, pivot = False, pivot_tol = 1e-4,
                        nest = None, mu = None, sigma = None,
                        spherical_basis_directory='./SphericalBasis',model_directory='./StanModels',output_directory='./StanOutput',
                        **kwargs):

        # Utilities
        self.order_to_nside = lambda order: 2**order
        self.nside_to_npix = lambda nside: 12*nside**2
        self.order_to_npix = lambda order: self.nside_to_npix(self.order_to_nside(order))

        self.spherical_basis_directory = self._verify_directory(spherical_basis_directory)
        self.model_directory = self._verify_directory(model_directory)
        self.output_directory = self._verify_directory(output_directory)

        self.sparse = sparse
        self.sparse_tol = sparse_tol
        self.pivot = pivot
        self.pivot_tol = pivot_tol
        self.nest = nest

        # Reshape k and n to be valid
        self._reshape_k_and_n(k,n,axes)

        # Process basis-specific options
        self._process_basis_options(**kwargs)

        # Load spherical basis
        self._load_spherical_basis()

        # Process covariance kernel options
        self.magnitude_bins = magnitude_bins if magnitude_bins is not None else np.arange(self.M+1)
        self.colour_bins = colour_bins if colour_bins is not None else np.arange(self.C+1)
        self.magnitude_kernel = magnitude_kernel if magnitude_kernel is not None else WhiteNoise(self.M)
        self.colour_kernel = colour_kernel if colour_kernel is not None else WhiteNoise(self.C)

        # Compute cholesky matrices
        self.M_subspace, self.cholesky_m = self._construct_cholesky_matrix(self.magnitude_kernel,self.magnitude_bins)
        self.C_subspace, self.cholesky_c = self._construct_cholesky_matrix(self.colour_kernel,self.colour_bins)

        # Process mu and sigma
        self._process_mu(mu)
        self._process_sigma(sigma)

        # Construct Stan Input
        self._construct_input()

        # File root
        self.file_root = file_root

    def stanoptimize(self, number_of_iterations = 1000, inits = 2):

        if not hasattr(self, 'stan_model'): self._load_stan_model()

        print('Running optimisation')
        t1 = time.time()
        _stan_optimum = self.stan_model.optimize(data = self.model_input, iter = number_of_iterations, output_dir = self.output_directory, inits = inits)
        t2 = time.time()
        print(f'Finished optimisation, it took {t2-t1:.1f} seconds')

        # Extract maxima
        _size_z = self.S*self.M_subspace*self.C_subspace
        _size_x = self.M*self.C*self.P
        self.optimum_lnp = _stan_optimum.optimized_params_np[0]
        self.optimum_z = np.transpose(_stan_optimum.optimized_params_np[1:1+_size_z].reshape((self.C_subspace,self.M_subspace,self.S)))
        self.optimum_x = np.transpose(_stan_optimum.optimized_params_np[1+_size_z:].reshape((self.P,self.C,self.M)))
        if self.nest:
            _ring_indices = hp.nest2ring(self.nside, np.arange(self.P))
            self.optimum_x = self._ring_to_nest(self.optimum_x)

        # Move convergence information somewhere useful
        import shutil
        self.optimum_convergence_file = self.file_root+'_convergence.txt'
        shutil.move(self.output_directory + str(_stan_optimum).split('/')[-1], self.output_directory + self.optimum_convergence_file)
        print(f'Convergence information stored in {self.output_directory + self.optimum_convergence_file}')

        self.optimum_results_file = self.file_root+'_results.h5'
        self.save_h5(t2-t1)

    def scipyoptimize(self, z=None, number_of_iterations=1000, ncores=2, bounds=None, method='BFGS',
                            force=False, nfev_init=0, verbose=0, **scipy_kwargs):

        from multiprocessing import Pool
        import scipy

        if not hasattr(self, 'scipy_args'): self._construct_scipy_input(sparse=self.sparse)
        self._parallelize_scipy_model(nsets=ncores)

        print('Initialising multiprocessing processes.')
        global evaluators
        evaluators = [evaluate(self.P_set[i], self.S, self.M, self.C, self.M_subspace, self.C_subspace,
                               self.scipy_args_set[i],
                               self.likelihood_function, j=self.j, verbose=verbose) for i in range(ncores)]
        evaluators[0].nfev=nfev_init

        self.optimum_results_file = self.file_root+'_scipy_results.h5'

        with Pool(ncores) as pool:
            icore = np.arange(ncores)

            def likelihood(z):

                evaluators[0].z = z
                evaluations = pool.map(evaluate_likelihood, zip(icore, np.repeat([z,],ncores, axis=0)))
                lnL, lnL_grad =  evaluators[0].merge_likelihoods(evaluations)

                return -lnL + 0.5*np.sum(z**2), -lnL_grad.flatten() + z
            if verbose>0: callback=evaluators[0].fcall
            else: callback=None

            print('Evaluate initial likelihood')
            if z is None: z = np.random.normal(size=(self.S, self.M_subspace, self.C_subspace))

            if number_of_iterations==0: return likelihood(z);
            else: likelihood(z); evaluators[0].fcall(z)

            import time
            print('Running optimisation')
            t1 = time.time()
            res = scipy.optimize.minimize(likelihood, z.flatten(), method=method, jac=True, bounds=bounds, callback=callback, **scipy_kwargs)
            t2 = time.time()
            print(f'Finished optimisation, it took {t2-t1:.1f} seconds')

            self.optimum_lnp_history = np.array(evaluators[0].lnlike_history)

            print('Processing results.')
            self.optimum_lnp = -res['fun']
            self.optimum_z = res['x'].reshape((self.S, self.M_subspace, self.C_subspace))
            self.optimum_x = self._get_x(self.optimum_z)
            if self.nest: self.optimum_x = self._ring_to_nest(np.moveaxis(self.optimum_x, 0, -1))

        self.optimum_results_file = self.file_root+'_results.h5'
        self.save_h5(t2-t1)

    def _get_likelihood(self, z, ncores=1):

        return self.scipyoptimize(z, ncores=ncores, number_of_iterations=0)


    def save_h5(self, runtime, *keys):

        # Save optimum to h5py
        with h5py.File(self.output_directory + self.optimum_results_file, 'w') as orf:
            orf.create_dataset('z', data = self.optimum_z, dtype = np.float64, compression = 'lzf', chunks = True)
            orf.create_dataset('x', data = self.optimum_x, dtype = np.float64, compression = 'lzf', chunks = True)
            orf.create_dataset('magnitude_bins', data = self.magnitude_bins, dtype = np.float64)
            orf.create_dataset('colour_bins', data = self.colour_bins, dtype = np.float64)
        print(f'Optimum values stored in {self.output_directory + self.optimum_results_file}')

    def sample(self, number_of_iterations=1000, number_of_warmups=100, threads=5, chains=4, inits=2):

        #_stan_samples = self.stan_model.sample(data=self.model_input, chains=chains, parallel_chains=chains, show_progress=True,
        #                                       iter_warmup=number_of_warmups, iter_sampling=number_of_iterations, output_dir=self.output_directory)

        _stan_samples = self.stan_model.sample(data=self.model_input, chains=chains, parallel_chains=chains, threads_per_chain=threads, show_progress=True, #refresh=10,
                                              iter_warmup=number_of_warmups, iter_sampling=number_of_iterations, output_dir=self.output_directory)

    def print_convergence(self, number_of_lines = 2):
        for line in self._tail(self.output_directory + self.optimum_convergence_file,number_of_lines):
            print(line)

    def _verify_directory(self,_directory):

        # Check it exists, if not then create
        if not os.path.exists(_directory):
            os.makedirs(_directory)

        # Ensure it ends with '/'
        if _directory[-1] != '/':
            _directory = _directory + '/'

        return _directory

    def _process_basis_options(self,options):
        pass

    def _reshape_k_and_n(self,k,n,axes):

        possible_axes = ["magnitude","colour","position"]
        axes_size = len(axes)

        assert 'position' in axes # We do not yet accept magnitude-colour only selection functions
        assert set(axes).issubset(set(possible_axes)) # We can only accept magnitude, colour and position
        assert axes_size == len(set(axes)) # Each axis must be unique
        assert k.shape == n.shape # k and n must have same shape

        new_indices = [possible_axes.index(axis) for axis in axes]
        self.k = np.moveaxis(k.copy().reshape(k.shape+(1,)*(3-axes_size)),range(axes_size),new_indices)
        self.n = np.moveaxis(n.copy().reshape(n.shape+(1,)*(3-axes_size)),range(axes_size),new_indices)

        self.M, self.C, self.P = self.k.shape
        if self.nest != None:
            assert hp.isnpixok(self.P) # number of pixels must be valid if we are using a healpix position basis

        if self.nest == True:
            self.k = self._nest_to_ring(self.k)
            self.n = self._nest_to_ring(self.n)

    def _load_spherical_basis(self):
        """ Loads in the spherical basis file. If they don't exist, then generate them. The generator must be implemented in each child class. """

        _spherical_basis_files = self.spherical_basis_file

        if not os.path.isfile(self.spherical_basis_directory + self.spherical_basis_file):
            print('Spherical basis file does not exist, generating... (this may take some time!)')
            self._generate_spherical_basis(self.spherical_basis_directory + self.spherical_basis_file)

        # Load spherical basis
        with h5py.File(self.spherical_basis_directory + self.spherical_basis_file, 'r') as sbf:
            self.basis = {k:v[()] for k,v in sbf.items()}

        print('Spherical basis file loaded')

    def _construct_cholesky_matrix(self, kernel, bins):

        # Compute cholesky
        _x = 0.5*(bins[1:]+bins[:-1])
        _N_subspace, _cholesky = kernel.cholesky(_x)

        return _N_subspace, _cholesky

    def _process_mu(self,mu):

        # Process mu
        if mu == None:
            self.mu = 0.0
        else:
            self.mu = mu

    def _process_sigma(self,sigma):

        # Process sigma
        if sigma == None:
            self.sigma = np.ones(self.S)
        elif isinstance(sigma, np.ndarray):
            assert sigma.shape == (self.S,)
            self.sigma = sigma
        elif callable(sigma):
            self.sigma = sigma(self.basis['modes'])
        elif type(sigma) in [list,tuple]:
            self.sigma = self._process_sigma_basis_specific(sigma)
        else:
            self.sigma = sigma*np.ones(self.S)

    def _process_mu_basis_specific(self,sigma):
        pass

    def _process_sigma_basis_specific(self,sigma):
        pass


    def _load_stan_model(self):

        _model_file = self.model_directory+f"{self.basis_keyword}_magnitude_colour_position{'_sparse' if self.sparse else ''}.stan"

        if not os.path.isfile(_model_file):
            with open(_model_file, 'w') as f:
                f.write(self._yield_stan_model())

        from cmdstanpy import CmdStanModel
        self.stan_model = CmdStanModel(stan_file = _model_file, cpp_options={"STAN_THREADS": True})

    def _yield_stan_model(self):
        pass

    def _construct_scipy_input(self):
        pass


    def _construct_input(self):

        self.model_input = {'k':self.k,
                           'n':self.n,
                           'P':self.P,
                           'M':self.M,
                           'M_subspace':self.M_subspace,
                           'C':self.C,
                           'C_subspace':self.C_subspace,
                           'S':self.S,
                           'mu':self.mu,
                           'sigma':self.sigma}

        # Add the basis specific keys
        self.model_input.update(self.basis)

        if self.sparse:
            self.model_input['cholesky_n_m'], self.model_input['cholesky_w_m'], self.model_input['cholesky_v_m'], self.model_input['cholesky_u_m'] = self._sparsify(self.cholesky_m)
            self.model_input['cholesky_n_c'], self.model_input['cholesky_w_c'], self.model_input['cholesky_v_c'], self.model_input['cholesky_u_c'] = self._sparsify(self.cholesky_c)
        else:
            self.model_input['cholesky_m'] = self.cholesky_m
            self.model_input['cholesky_c'] = self.cholesky_c

        integer_types = (np.int8, np.int16, np.int32, np.int64, np.uint8, np.uint16, np.uint32, np.uint64)
        for k,v in self.model_input.items():

            # Convert everything to numpy array
            if isinstance(v, (list, tuple)):
                v = np.array(v)

            # Correct for Stan being one-indexed - only apply to one-dimensional collections of integers
            if isinstance(v, np.ndarray) and v.dtype in integer_types:
                if k not in ['n','k','modes']:
                    v = v + 1

            # Deal with serialisation issue with integers
            if isinstance(v, integer_types):
                v = v.item()

            self.model_input[k] = v



    def _sparsify(self,_matrix):

        # Set any elements in each row that are smaller than self.sparse_tol * max(row) to zero
        _height,_width = _matrix.shape
        _sparse_matrix = _matrix.copy()

        for n in range(_height):
            _row = np.abs(_matrix[n])
            _sparse_matrix[n,_row/max(_row) < self.sparse_tol] = 0

        # Compute the CSR decomposition of the sparse matrix
        from scipy.sparse import csr_matrix
        _csr_matrix = csr_matrix(_sparse_matrix)
        _csr_n = _csr_matrix.data.size
        print(f"{100*(1.0-_csr_n/(_height*_width)):.2f}% sparsity")

        return _csr_n, _csr_matrix.data, _csr_matrix.indices, _csr_matrix.indptr


    def _tail(self,filename, lines=1, _buffer=4098):
        """Tail a file and get X lines from the end"""
        # place holder for the lines found
        lines_found = []

        # block counter will be multiplied by buffer
        # to get the block size from the end
        block_counter = -1

        with open(filename,'r') as f:
            # loop until we find X lines
            while len(lines_found) < lines:
                try:
                    f.seek(block_counter * _buffer, os.SEEK_END)
                except IOError:  # either file is too small, or too many lines requested
                    f.seek(0)
                    lines_found = f.readlines()
                    break

                lines_found = f.readlines()

                block_counter -= 1

        return lines_found[-lines:]

    def _nest_to_ring(self,A):
        """ Reorders an array of shape (M,C,P) to ring ordering. """
        _npix = A.shape[2]
        _nside = hp.npix2nside(_npix)
        _reordering = hp.ring2nest(_nside, np.arange(_npix))
        return A[:,:,_reordering]

    def _ring_to_nest(self,A):
        """ Reorders an array of shape (M,C,P) to ring ordering. """
        _npix = A.shape[2]
        _nside = hp.npix2nside(_npix)
        _reordering = hp.nest2ring(_nside, np.arange(_npix))
        return A[:,:,_reordering]

    def _generate_spherical_basis(self,gsb_file):
        pass



def evaluate_likelihood(iz):
    return evaluators[iz[0]].evaluate_likelihood(iz[1])
def evaluate_x(iz):
    return iz[0], evaluators[iz[0]].evaluate_x(iz[1])

class evaluate():

    def __init__(self, P, S, M, C, M_subspace, C_subspace, scipy_args, likelihood_function, j=[-1], verbose=0):

        self.P=P
        self.S=S
        self.M=M
        self.C=C
        self.M_subspace=M_subspace
        self.C_subspace=C_subspace
        self.scipy_args=scipy_args

        self.likelihood_function = likelihood_function

        self.j=j

        self.x = np.zeros((self.P, self.M, self.C))
        self.lnL_grad = np.zeros((self.S, self.M_subspace, self.C_subspace))

        self.tinit = time.time()
        self.lnlike_iter = 0.
        self.gnorm_iter = 0.
        self.nfev = 0
        self.lnlike_history = []
        self.lnlike_smhistory = []
        self.z_prev = np.zeros((self.S,self.M_subspace,self.C_subspace))

        self.verbose=verbose

    def evaluate_likelihood(self, z):

        x = np.zeros((self.P, self.M, self.C))
        lnL_grad = np.zeros((self.S, self.M_subspace, self.C_subspace))

        lnL, lnL_grad = self.likelihood_function(z.reshape((self.S, self.M_subspace, self.C_subspace)), self.M, self.C, self.P, *self.scipy_args, x, lnL_grad)

        return lnL, lnL_grad

    def merge_likelihoods(self, evaluations):

        lnL=0.
        grad = np.zeros((self.S, self.M_subspace, self.C_subspace))

        for e in evaluations:
            lnL += e[0]
            grad += e[1]

        self.lnlike_iter = lnL
        self.gnorm_iter = np.sqrt(np.sum(grad**2))/(self.S * self.M_subspace * self.C_subspace)
        self.nfev += 1
        self.lnlike_history.append(lnL)

        return lnL, grad.flatten()

    def fcall(self, X):

        if self.nfev%self.verbose==0:
            self.lnlike_smhistory.append(self.lnlike_iter)

            i=0; std_str=""
            for j in self.j:
                if j==-1: std_str+=f"_{np.std(X.reshape(self.S, self.M_subspace, self.C_subspace)[i]):.2f}"
                else: std_str+=f"_{np.std(X.reshape(self.S, self.M_subspace, self.C_subspace)[i:int(i+hp.nside2npix(pow(2,j))+0.1)]):.2f}"
                i+=int(hp.nside2npix(pow(2,j))+0.1)

            print(f't={int(time.time()-self.tinit):03d}, n={self.nfev:02d}, lnL={self.lnlike_iter:.0f}, gnorm={self.gnorm_iter:.5f}')

            self.z_prev=X.copy()
            self.lnL_prev=self.lnlike_iter



#!/usr/bin/env python
#
# hammer.py
# Uses spherical harmonics to estimate a selection function across the sky.
#
# Copyright (C) 2021  Douglas Boubert & Andrew Everall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import numpy as np
import healpy as hp
import tqdm
import h5py
from .base import Base
import os

class Hammer(Base):

    basis_keyword = 'harmonic'

    def _process_basis_options(self,lmax = 0):
        self.lmax = lmax
        self.S = (self.lmax + 1) ** 2
        self.nside = hp.npix2nside(self.P)
        self.spherical_basis_file = f'{self.basis_keyword}_nside{self.nside}_lmax{self.lmax}.h5'

    def _process_sigma_basis_specific(self,sigma):
        assert len(sigma) == 2
        return np.sqrt(np.exp(sigma[0])*np.power(1.0+self.basis['modes'],sigma[1]))

    def _generate_spherical_basis(self,gsb_file):

        nside = self.nside
        lmax = self.lmax
        Npix = self.P

        # Form the l's and m's
        Nmodes = int((lmax+1)**2)
        Nmodes_hp = int((lmax+1)*(lmax+2)/2)
        l_hp,m_hp = hp.sphtfunc.Alm.getlm(lmax=lmax)
        assert Nmodes_hp == l_hp.size

        l, m = np.zeros(Nmodes,dtype=int), np.zeros(Nmodes,dtype=int)
        l[:Nmodes_hp],m[:Nmodes_hp] = l_hp,m_hp
        l[Nmodes_hp:],m[Nmodes_hp:] = l_hp[lmax+1:],-m_hp[lmax+1:]

        # Ring idxs of pixels with phi=0
        theta, phi = hp.pix2ang(nside, np.arange(hp.nside2npix(nside)))
        theta_ring, unique_idx, jpix = np.unique(theta, return_index=True, return_inverse=True)

        # Generate lambda
        _lambda = np.zeros((Nmodes, 4*nside-1))
        if False: # From scipy
            # For |m|>0 this comes out a factor of 2 smaller than the healpy version
            # For m<0 there's also a factor of (-1)^m difference
            for i,(_l,_m) in enumerate(zip(tqdm.tqdm(l),m)):
                _lambda[i] = (-1)**np.abs(_m) * np.real( scipy.special.sph_harm(np.abs(_m), _l, theta_ring*0., theta_ring) )
        else: # From healpy
            alm_hp = np.zeros(Nmodes_hp)
            for i,(_l,_m) in enumerate(zip(tqdm.tqdm(l),m)):
                i_hp = hp.sphtfunc.Alm.getidx(lmax, _l, np.abs(_m))
                alm_hp = np.zeros(Nmodes_hp)*(0.+0.j)
                # Get real component
                alm_hp[i_hp] = 1.+0.j
                map_hp = (1.+0.j)*hp.sphtfunc.alm2map(alm_hp,nside=nside, verbose=False)
                # Add imaginary component
                alm_hp[i_hp] = 0.+1.j
                map_hp += (0.-1.j)*hp.sphtfunc.alm2map(alm_hp,nside=nside, verbose=False)
                alm_hp[i_hp] = 0.+0.j
                map_hp /= np.exp(1.j*np.abs(_m)*phi)
                # Select unique latitude indices
                _lambda[i] = (-1)**np.abs(_m) * np.real(map_hp)[unique_idx]

                # Divide by 2
                if _m != 0:
                    _lambda[i] /= 2.0

        # Generate Exponential
        azimuth = np.ones((2*lmax+1,Npix))
        for _m in range(-lmax, lmax+1):
            if _m<0:   azimuth[_m+lmax] = np.sqrt(2) * np.sin(-_m*phi)
            elif _m>0: azimuth[_m+lmax] = np.sqrt(2) * np.cos(_m*phi)
            else: pass

        # Generate indices mapping m to alm
        lower, upper = np.zeros(2*lmax+1),np.zeros(2*lmax+1)
        for i, _m in enumerate(range(-lmax,lmax+1)):
            match = np.where(m==_m)[0]
            lower[i] = match[0]
            upper[i] = match[-1]

        save_kwargs = {'compression':"lzf", 'chunks':True, 'fletcher32':False, 'shuffle':True}
        with h5py.File(gsb_file, 'w') as f:
            # Create datasets
            f.create_dataset('lambda', data = _lambda.T, shape = (4*nside-1, Nmodes,), dtype = np.float64, **save_kwargs)
            f.create_dataset('azimuth',data = azimuth, shape = (2*lmax+1, Npix, ),   dtype = np.float64, **save_kwargs)
            f.create_dataset('modes',      data = l,       shape = (Nmodes,), dtype = np.uint32, scaleoffset=0, **save_kwargs)
            f.create_dataset('submodes',      data = m,       shape = (Nmodes,), dtype = np.int32, scaleoffset=0, **save_kwargs)
            f.create_dataset('pixel_to_ring',   data = jpix.astype(int),    shape = (Npix,),   dtype = np.uint32, scaleoffset=0, **save_kwargs)
            f.create_dataset('lower',   data = lower.astype(int),    shape = (2*lmax+1,),   dtype = np.uint32, scaleoffset=0, **save_kwargs)
            f.create_dataset('upper',   data = upper.astype(int),    shape = (2*lmax+1,),   dtype = np.uint32, scaleoffset=0, **save_kwargs)
            f.create_dataset('L', data = 2 * self.lmax + 1)
            f.create_dataset('R', data = 4 * self.nside - 1)

    def _yield_stan_model(self):
        if self.sparse:
            cholesky_parameters = '''
            int cholesky_n_m;                     // sparse cholesky in magnitude - number of nonzero elements
            row_vector[cholesky_n_m] cholesky_w_m;// sparse cholesky in magnitude - nonzero elements
            int cholesky_v_m[cholesky_n_m];       // sparse cholesky in magnitude - columns of nonzero elements
            int cholesky_u_m[M+1];                // sparse cholesky in magnitude - where in w each row starts
            int cholesky_n_c;                     // sparse cholesky in colour - number of nonzero elements
            vector[cholesky_n_c] cholesky_w_c;    // sparse cholesky in colour - nonzero elements
            int cholesky_v_c[cholesky_n_c];       // sparse cholesky in colour - columns of nonzero elements
            int cholesky_u_c[C+1];                // sparse cholesky in colour - where in w each row starts
            '''
            cholesky_loop = '''
            a[s] = sigma[s] * (cholesky_w_m[cholesky_u_m[m]:cholesky_u_m[m+1]-1] * z[s,cholesky_v_m[cholesky_u_m[m]:cholesky_u_m[m+1]-1], cholesky_v_c[cholesky_u_c[c]:cholesky_u_c[c+1]-1]] * cholesky_w_c[cholesky_u_c[c]:cholesky_u_c[c+1]-1]);
            '''
        else:
            cholesky_parameters = '''
            row_vector[M_subspace] cholesky_m[M]; // Cholesky factor in magnitude space
            vector[C_subspace] cholesky_c[C];     // Cholesky factor in colour space
            '''
            cholesky_loop = '''
            a[s] = sigma[s] * cholesky_m[m] * z[s] * cholesky_c[c];
            '''

        stan_model = f'''
        data {{
            int<lower=0> P;                       // number of pixels
            int<lower=0> M;                       // number of bins in magnitude space
            int<lower=0> M_subspace;              // number of inducing points in magnitude space
            int<lower=0> C;                       // number of bins in colour space
            int<lower=0> C_subspace;              // number of inducing points in colour space
            int<lower=0> L;                       // 2 * max l of hamonics + 1
            int<lower=0> S;                       // number of harmonics
            int<lower=0> R;                       // number of HEALPix isolatitude rings
            matrix[R,S] lambda;                   // spherical harmonics decomposed
            matrix[L,P] azimuth;                  // spherical harmonics decomposed
            int pixel_to_ring[P];                 // map P->R
            int lower[L];                         // map S->L
            int upper[L];                         // map S->L
            real mu;                              // mean across sky
            vector[S] sigma;                      // sigma of each harmonic
            int k[M,C,P];                         // number of heads
            int n[M,C,P];                         // number of flips
            {cholesky_parameters}
        }}
        parameters {{
            matrix[M_subspace,C_subspace] z[S];
        }}
        transformed parameters {{

            vector[P] x[M,C]; // Probability in logit-space
                
            // Loop over magnitude and colour
            for (m in 1:M){{
                for (c in 1:C){{
                    
                    // Local variables
                    vector[S] a;
                    matrix[R,L] F;

                    // Compute a 
                    for (s in 1:S){{
                        {cholesky_loop}
                    }}

                    // Compute F
                    for (l in 1:L) {{
                        F[:,l] = lambda[:,lower[l]:upper[l]] * a[lower[l]:upper[l]];
                    }}

                    // Compute x
                    for (p in 1:P){{
                        x[m,c,p] = mu + dot_product(F[pixel_to_ring[p]],azimuth[:,p]);
                    }}

                }}  
            }}

        }}
        model {{

            // Prior
            for (s in 1:S){{
                to_vector(z[s]) ~ std_normal();
            }}
            
            // Likelihood
            for (m in 1:M){{
                for (c in 1:C){{
                    k[m,c] ~ binomial_logit(n[m,c], x[m,c]);
                }}
            }}
            
        }}
        '''

        return stan_model



#!/usr/bin/env python
#
# setup.py
# Package "selectionfunctiontoolbox" for pip.
#
# Copyright (C) 2021  Douglas Boubert & Andrew Everall
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import pathlib
from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent

VERSION = '1.0.0'
PACKAGE_NAME = 'selectionfunctiontoolbox'
AUTHOR = 'Douglas Boubert & Andrew Everall'
AUTHOR_EMAIL = 'ask@gaiaverse.space'
URL = 'https://github.com/gaiaverse/selectionfunctiontoolbox'

LICENSE = 'GPLv2'
DESCRIPTION = 'A set of tools for estimating the selection function of a scientific sample drawn from an astronomical catalogue.'
LONG_DESCRIPTION = (HERE / "README.md").read_text()
LONG_DESC_TYPE = "text/markdown"

INSTALL_REQUIRES = [
        'numpy',
        'scipy',
        'astropy',
        'h5py',
        'healpy',
]

setup(name=PACKAGE_NAME,
      version=VERSION,
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      long_description_content_type=LONG_DESC_TYPE,
      author=AUTHOR,
      license=LICENSE,
      author_email=AUTHOR_EMAIL,
      url=URL,
      download_url = 'https://github.com/gaiaverse/selectionfunctiontoolbox/archive/v1.0.0.tar.gz',
      install_requires=INSTALL_REQUIRES,
      packages=find_packages()
      )